//============================================================================
// Name        : CoDi.h
// Author      : D.A.Klein
// Version     : 1.1
// Copyright   : (C) University of Bonn
// Description : CoDi: Continuous-Distributions Saliency
// License     : GNU GPLv3 (see http://www.gnu.org/copyleft/gpl.html)
//============================================================================

/*
If you use this software or parts of it within scientific work, please cite
the original paper it was introduced:

@InProceedings{klein_dagm12,
  author = {Dominik A. Klein and Simone Frintrop},
  title = {{Salient Pattern Detection using $W_2$ on Multivariate Normal Distributions}},
  booktitle = {Proc.~of DAGM and OAGM},
  year = {2012},
  month = {Aug.},
  address = {Graz, Austria}
}
*/

#ifndef CoDi_H_
#define CoDi_H_

#include <opencv2/opencv.hpp>

#include <fstream>
#include <iostream>

#include <vector>
#include <algorithm>
#include <stdlib.h>

#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>

#include <boost/serialization/nvp.hpp>
#include <boost/serialization/export.hpp>
#include <boost/serialization/extended_type_info.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/assume_abstract.hpp>

#include "GPyr.h"
#include "ImageOperations.h"
#include "CoDicfg.h"

#define FILTER_PRECISION 4.0

namespace cv
{

template <typename _Tp> class CoDi
{
public:
	// constructor without setting special parameters
	CoDi() {};
	// constructor that adopts special config settings
	CoDi(const CoDicfg& config): cfg(config) {};
	virtual ~CoDi() {};

	// set a new config
	void setCfg(const CoDicfg& config) { cfg = config; }

	// read the actual config
	const CoDicfg& getCfg() { return cfg; }

	// compute saliency from an image
	void computeFrom(const cv::Mat_<cv::Vec<_Tp, 3> >& rgb) {
		// convert color space
		cv::Mat_<cv::Vec<_Tp, 3> > img;
		this->convertColor(cfg.color_space, rgb, img);

		if (cfg.maxOctaves) {
			// adjust number of octaves to input image resolution
			cfg.octaves = std::max(1, int(log2(std::min(img.rows, img.cols)/(cfg.sizeIn*3.0))));
			std::cout << "Number of octaves set to " << cfg.octaves << "\n";
		}

		// compute Gaussian pyramid
		cv::GPyr_<cv::Vec<_Tp, 3> > pyr(cfg.scales, cfg.octaves, 1, cfg.preSmoothing, img, FILTER_PRECISION);

		// some temporary Mats needed
		cv::Mat_<cv::Vec<_Tp, 3> > scale_img;
		cv::Mat_<cv::Vec<_Tp, 2> > dxy, cor_1d, incor_1d, outcor_1d;
		cv::Mat_<cv::Vec<_Tp, 5> > cor_2d, incor_2d, outcor_2d;
		cv::Mat_<_Tp> tmp, tmp2;
		cv::Vec<_Tp, 2> avg_1d;
		cv::Vec<_Tp, 5> avg_2d;

		// resize vectors of result Mats
		intensities.resize(cfg.octaves);
		colors.resize(cfg.octaves);

		// initialize conspicuity and saliency maps
		this->initMap(cfg.scale_fusion, img.size(), intensity);
		this->initMap(cfg.scale_fusion, img.size(), color);
		this->initMap(cfg.feature_fusion, img.size(), saliency);

		// for every octave
		for (int o=0; o<cfg.octaves; ++o) {
			this->initMap(cfg.scale_fusion, pyr(0,o).size(), intensities[o]);
			this->initMap(cfg.scale_fusion, pyr(0,o).size(), colors[o]);

			// for every scale
			for (int s=0; s<cfg.scales; ++s) {

				// difference-of-Gaussian scales
				scale_img = pyr(s-1,o) - pyr(s,o);
				scale_img *= pyr.sigmaScale(s-1,o);

				// filter sizes corrected by scale factor
				const double sizeIn = cfg.sizeIn * std::pow(2.0, (double)s/(double)cfg.scales);
				const double sizeOut = cfg.sizeOut * std::pow(2.0, (double)s/(double)cfg.scales);

				// intensity
				if (cfg.intensity) {
					cv::correlateIntensity(scale_img, cor_1d);
					integrate(cfg.integrator, cor_1d, sizeIn, sizeOut, incor_1d, outcor_1d);
					distance(cfg.type, incor_1d, outcor_1d, tmp);
					// combine scales
					this->fuseMaps(cfg.scale_fusion, tmp, intensities[o]);
				}

				// color
				if (cfg.color) {
					cv::correlateColor(scale_img, cor_2d);
					integrate(cfg.integrator, cor_2d, sizeIn, sizeOut, incor_2d, outcor_2d);
					distance(cfg.type, incor_2d, outcor_2d, tmp);
					// combine scales
					this->fuseMaps(cfg.scale_fusion, tmp, colors[o]);
				}

			}
			// combine octaves
			cv::resize(intensities[o], tmp, intensity.size(), 0.0, 0.0, cv::INTER_CUBIC);
			this->fuseMaps(cfg.scale_fusion, tmp, intensity);
			cv::resize(colors[o], tmp, color.size(), 0.0, 0.0, cv::INTER_CUBIC);
			this->fuseMaps(cfg.scale_fusion, tmp, color);
		}

		// normalize by the number of scales*octaves
		this->normMap(cfg.scale_fusion, cfg.scales*cfg.octaves, intensity);
		this->normMap(cfg.scale_fusion, cfg.scales*cfg.octaves, color);

		// fuse feature channels
		int n=0;
		if (cfg.intensity) { this->fuseMaps(cfg.feature_fusion, intensity, saliency); ++n;}
		if (cfg.color) { this->fuseMaps(cfg.feature_fusion, color, saliency); ++n; }

		// normalize saliency by number of feature channels
		this->normMap(cfg.feature_fusion, n, saliency);
	}

	// saliency map from last computeFrom() call
	const cv::Mat_<_Tp>& getSalMap() const { return saliency; }

	// conspicuity maps from last computeFrom() call
	const cv::Mat_<_Tp>& getIntensityConsp() const { return intensity; }
	const cv::Mat_<_Tp>& getColorConsp() const { return color; }

	// conspicuity pyramids from last computeFrom() call
	const std::vector<cv::Mat_<_Tp> >& getIntensities() const { return intensities; }
	const std::vector<cv::Mat_<_Tp> >& getColors() const { return colors; }

protected:
	// converts color from rgb to desired color space
	void convertColor(const colorSpace cs, const cv::Mat_<cv::Vec<_Tp, 3> >& src, cv::Mat_<cv::Vec<_Tp, 3> >& dst) const {
		switch(cs) {
			case IPP: cv::cvtRGB2IPP(src, dst, false); break;
			case IPP_OPT: cv::cvtRGB2IPP(src, dst, true); break;
			case IAB: cv::cvtRGB2IAB(src, dst); break;
			case IXY: cv::cvtRGB2IXY(src, dst); break;
			case IOP: cv::cvtRGB2IOp(src, dst); break;
			case IOP_NORM: cv::cvtRGB2IOpNorm(src, dst); break;
			case LAB: cv::cvtRGB2Lab(src, dst); cv::multiply(dst, cv::Vec<_Tp,3>(std::sqrt(2.0), 1.0, 1.0), dst); break;
		}
	}

	// initializes the mats with the identity element of the fuse operation
	void initMap(const fuseOp op, const cv::Size2i& size, cv::Mat_<_Tp>& mat) const {
		switch(op) {
			case FUSE_AM: case FUSE_MAX: case FUSE_HM: case FUSE_QM: { mat = cv::Mat_<_Tp>::zeros(size); break; }
			case FUSE_GM: case FUSE_GM2: { mat = cv::Mat_<_Tp>::ones(size); break; }
		}
	}

	// fuses two maps without normalization step (one can chain fusion steps and
	// normalize in the end).
	void fuseMaps(const fuseOp op, const cv::Mat_<_Tp>& src, cv::Mat_<_Tp>& dst) const {
		CV_Assert(src.cols == dst.cols && src.rows == dst.rows);
		switch(op) {
			case FUSE_AM: { cv::add(dst, src, dst); break; }
			case FUSE_GM: { cv::multiply(dst, src, dst); break; }
			case FUSE_HM: { dst = dst + (1.0 / src); break; }
			case FUSE_QM: { dst += src.mul(src); break; }
			case FUSE_MAX: { dst = cv::max(dst, src); break; }
			case FUSE_GM2: { cv::Mat_<_Tp> tmp = (src + cv::Mat_<_Tp>::ones(src.size())*0.01); cv::multiply(dst, tmp, dst); break; }
		}
	}

	// normalize maps according to the number of maps that had been fused
	void normMap(const fuseOp op, const int n, cv::Mat_<_Tp>& mat) const {
		switch(op) {
			case FUSE_AM: { mat *= (1.0/n); break; }
			case FUSE_GM: case FUSE_GM2: { cv::pow(mat, (1.0/n), mat); break; }
			case FUSE_HM: mat = ((_Tp)n) / mat; break;
			case FUSE_QM: mat *= (1.0/n); cv::sqrt(mat, mat); break;
			case FUSE_MAX: { break; }
		}
	}

	// apply an integration window for center and surround
	template <int n> void integrate(const CSintegration integr, const cv::Mat_<cv::Vec<_Tp, n> >& src, const double& r_c, const double& r_s, cv::Mat_<cv::Vec<_Tp, n> >& center, cv::Mat_<cv::Vec<_Tp, n> >& surround) const {
		CV_Assert(r_c < r_s);
		switch (integr) {
			case CS_BOX: cv::filterBox(src, r_c, r_s, center, surround); break;
			case CS_GAUSSIAN: cv::filterGaussian(src, r_c, r_s, center, surround, FILTER_PRECISION); break;
			case CS_BOX_GAUSSIAN: cv::filterBoxGaussian(src, r_c, r_s, center, surround, FILTER_PRECISION); break;
			case CS_DOG: cv::filterDoG(src, r_c, r_s, center, surround, FILTER_PRECISION); break;
		}
	}

	// compute center-surround distance for a 1-dimensional feature
	void distance(const distrFeat type, cv::Mat_<cv::Vec<_Tp, 2> >& incor, cv::Mat_<cv::Vec<_Tp, 2> >& outcor, cv::Mat_<_Tp>& result) const {
		switch (type) {
			case DIFF_ENTROPY: cv::entropy1D(incor, result); break;
			case KLD: cv::kld1D(incor, outcor, result); break;
			case HELLINGER: cv::hellinger1D(incor, outcor, result); break;
			case FRECHET: cv::frechet1D(incor, outcor, result); break;
		}
	}

	// compute center-surround distance for a 2-dimensional feature
	void distance(const distrFeat type, cv::Mat_<cv::Vec<_Tp, 5> >& incor, cv::Mat_<cv::Vec<_Tp, 5> >& outcor, cv::Mat_<_Tp>& result) const {
		switch (type) {
			case DIFF_ENTROPY: cv::entropy2D(incor, result); break;
			case KLD: cv::kld2D(incor, outcor, result); break;
			case HELLINGER: cv::hellinger2D(incor, outcor, result); break;
			case FRECHET: cv::frechet2D(incor, outcor, result); break;
		}
	}

	// the parameters of the algorithm
	CoDicfg cfg;

	// (intermediate) result mats
	std::vector<cv::Mat_<_Tp> > intensities;
	std::vector<cv::Mat_<_Tp> > colors;

	cv::Mat_<_Tp> intensity;
	cv::Mat_<_Tp> color;

	cv::Mat_<_Tp> saliency;

};

} /* namespace cv */
#endif /* CoDi_H_ */
