//============================================================================
// Name        : CoDicfg.h
// Author      : D.A.Klein
// Version     : 1.1
// Copyright   : (C) University of Bonn
// Description : CoDi: Continuous-Distributions Saliency
// License     : GNU GPLv3 (see http://www.gnu.org/copyleft/gpl.html)
//============================================================================

/*
If you use this software or parts of it within scientific work, please cite
the original paper it was introduced:

@InProceedings{klein_dagm12,
  author = {Dominik A. Klein and Simone Frintrop},
  title = {{Salient Pattern Detection using $W_2$ on Multivariate Normal Distributions}},
  booktitle = {Proc.~of DAGM and OAGM},
  year = {2012},
  month = {Aug.},
  address = {Graz, Austria}
}
*/

#ifndef CoDiCFG_H_
#define CoDiCFG_H_


namespace cv
{

enum colorSpace {
	// "Color Texture Classiﬁcation by Normalized Color Space Representation", by Constantin Vertan, Nozha Boujemaa
	// without selection of best rotation of color plane
	IPP=0,
	// with color rotation according to colors appearing in the image
	IPP_OPT=1,
	// Intensity, alpha, beta
	// http://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma
	IAB=2,
	// Intensity, hue/saturation in Cartesian X-Y
	IXY=3,
	// Intensity, Opponent colors (Red<->Green , Blue<->Yellow)
	IOP=4,
	// Intensity, Opponent colors (Red<->Green , Blue<->Yellow), normalized intensity
	IOP_NORM=5,
	// L*a*b*
	// http://en.wikipedia.org/wiki/Lab_color_space
	LAB=6
};

enum CSintegration {
	// Box filter
	CS_BOX=0,
	// Gaussian filter (slow for huge radii)
	CS_GAUSSIAN=1,
	// Fast, but precise Gaussian approximation using box filters
	CS_BOX_GAUSSIAN=2,
	// Difference of Gaussians (slow for huge radii)
	CS_DOG=3
};

enum distrFeat {
	// differential entropy
	DIFF_ENTROPY=0,
	// Kullback-Leibler divergence (relative entropy)
	KLD=1,
	// Hellinger distance
	HELLINGER=2,
	// Fréchet distance (Wasserstein metric on L2)
	FRECHET=3
};

enum fuseOp {
	// arithmetic mean
	FUSE_AM=0,
	// geometric mean
	FUSE_GM=1,
	// maximum
	FUSE_MAX=2,
	// geometric mean from maps+epsilon
	FUSE_GM2=3,
	// harmonic mean
	FUSE_HM=4,
	// root mean square (quadratic mean)
	FUSE_QM=5 };

/*
 * A configuration for the CoDi-Saliency algorithm. This class supports easy file input/output
 * via boost::serialization.
 */
class CoDicfg {
public:
	/*
	 * Constructor
	 * It sets some standart values hopefully reasonable for many applications. However,
	 * please check if you need to adapt values for your task.
	 */
	CoDicfg():
		scales(2), octaves(6), maxOctaves(false), preSmoothing(0.0),
		integrator(CS_BOX_GAUSSIAN), sizeIn(1.0), sizeOut(10.0),
		intensity(true), color(true), color_space(IOP),
		type(FRECHET), scale_fusion(FUSE_AM), feature_fusion(FUSE_AM) {};
	// Number of scales per octave for the Gaussian pyramid
	int scales;
	// Number of octaves for construction of the Gaussian pyramid
	int octaves;
	// adjust the number of octaves to fit the resolution of the input image. If this is set true,
	// it overrides the values of "octaves".
	bool maxOctaves;
	// Sigma to smooth the input image with before calculating saliency. This should
	// be set to 0 if one wants to use the original image
	double preSmoothing;
	// method to integrate local statistics
	CSintegration integrator;
	// Parameter for the center feature integration window
	double sizeIn;
	// Parameter for the surround feature integration window
	double sizeOut;
	// use intensity feature
	bool intensity;
	// use color feature
	bool color;
	// which color space to use
	colorSpace color_space;
	// method to compute center-surround difference
	distrFeat type;
	// method to fuse different scales of the same feature type
	fuseOp scale_fusion;
	// method to fuse different feature types
	fuseOp feature_fusion;

private:
    friend class boost::serialization::access;
    template<class Archive>
    /**
     * @param[in] ar
     * @param[in] version
     */
    void serialize(Archive & ar, const unsigned int version)
    {
    	ar & BOOST_SERIALIZATION_NVP(scales);
    	ar & BOOST_SERIALIZATION_NVP(octaves);
    	ar & BOOST_SERIALIZATION_NVP(maxOctaves);
    	ar & BOOST_SERIALIZATION_NVP(preSmoothing);
    	ar & BOOST_SERIALIZATION_NVP(integrator);
    	ar & BOOST_SERIALIZATION_NVP(sizeIn);
    	ar & BOOST_SERIALIZATION_NVP(sizeOut);
    	ar & BOOST_SERIALIZATION_NVP(intensity);
    	ar & BOOST_SERIALIZATION_NVP(color);
    	ar & BOOST_SERIALIZATION_NVP(color_space);
    	ar & BOOST_SERIALIZATION_NVP(type);
    	ar & BOOST_SERIALIZATION_NVP(scale_fusion);
    	ar & BOOST_SERIALIZATION_NVP(feature_fusion);
    }
};

} /* namespace cv */

#endif /* CoDiCFG_H_ */
