//============================================================================
// Name        : GPyr.h
// Author      : D.A.Klein
// Version     : 1.1
// Copyright   : (C) University of Bonn
// Description : CoDi: Continuous-Distributions Saliency
// License     : GNU GPLv3 (see http://www.gnu.org/copyleft/gpl.html)
//============================================================================

/*
If you use this software or parts of it within scientific work, please cite
the original paper it was introduced:

@InProceedings{klein_dagm12,
  author = {Dominik A. Klein and Simone Frintrop},
  title = {{Salient Pattern Detection using $W_2$ on Multivariate Normal Distributions}},
  booktitle = {Proc.~of DAGM and OAGM},
  year = {2012},
  month = {Aug.},
  address = {Graz, Austria}
}
*/

#ifndef GPYR_H_
#define GPYR_H_

#include <opencv2/opencv.hpp>
#include <vector>
#include <math.h>


namespace cv
{

/*
 * class GPyr_<_Tp>
 * Implements a Gaussian pyramid construction. The number of octaves and scales per octave can be choosen.
 */
template <typename _Tp> class GPyr_
{
public:
	/*
	* Constructor
	* @param[in]
	* sc	Number of scales per octave.
	* oct	Number of octaves. An octave is a scale step of doubling sigma. From one octave to the next,
	* 		horizontal and vertical resolution is halfed.
	* oh	Overhang in each octave. For some algorithms (e.g. DoG), one needs additional neighboring
	* 		scales computed at the same resolution.
	* sB	Sigma value the original image is smoothed with before calculating the pyramid.
	* img	The input image from which the pyramid is constructed.
	* fP	Filter precision for the gaussian smoothing. The filter radius is fP x sigma.
	*/
	GPyr_(const int sc, const int oct, const int oh, const double& sB, const cv::Mat_<_Tp>& img, const double fP=3.0):
		scales(sc), octaves(oct), overhang(oh), sigmaBegin(sB), filterPrecision(fP), pyramid(oct, std::vector<cv::Mat_<_Tp> >(sc+2*oh, cv::Mat_<_Tp>()))
	{
		CV_Assert(sc>0 && oct>0 && oh>=0 && sB>=0.0);

		this->compute(img);
	}

	virtual ~GPyr_() {};

	/*
	 * Access operator
	 * @param[in]
	 * sc	Scale to access. Available scales are from -overhang to scales+overhang-1.
	 * oct	Octave to access.
	 * @param[out]
	 * Reference to the Mat_ representing the selected scale.
	 */
	cv::Mat_<_Tp>& operator () (const int sc, const int oct) {
		CV_Assert(-overhang<=sc && 0<=oct && sc<scales+overhang && oct<octaves);

		return pyramid[oct][sc+overhang];
	}

	/*
	 * Computes the corresponding sigma for a certain scale of the pyramid.
	 * @param[in]
	 * sc	The scale for which the sigma is computed.
	 * oct	The octave.
	 */
	double sigmaScale(const double sc, const int oct=0) const {
		return std::pow(2.0, sigmaBegin + oct + ((double)(sc+overhang)/(double)scales));
	}

	/*
	 * Calculates the sigma that is needed for additional Gaussian smoothing, if one wants to reach
	 * a intermediate scale.
	 * @param[in]
	 * sc	The scale from which one wants to start.
	 * diff	The additional scale step one wants to smooth given in relation to a regular scale step of
	 * 		this pyramid.
	 * @param[out]
	 * The sigma one should additionaly smooth scale sc with.
	 */
	double sigmaDiff(const int sc, const double diff) const {
		const double s1 = this->sigmaScale(sc, 0);
		const double s2 = this->sigmaScale(sc+diff, 0);
		return std::sqrt(s2 * s2 - s1 * s1);
	}

protected:
//	cv::Mat_<_Tp>& operator () (const int sc, const int oct) {
//		CV_Assert(-overhang<=sc && 0<=oct && sc<scales+overhang && oct<octaves);
//
//		return pyramid[oct][sc+overhang];
//	}

	void compute(const cv::Mat_<_Tp>& img) {
		// copy the original image
		img.copyTo((*this)(-overhang,0));

		int filterSize;
		// smoothing the first image
		if (sigmaBegin > 0.0) {
			filterSize = 2*(int)(filterPrecision*sigmaBegin) + 1;
			cv::GaussianBlur((*this)(-overhang,0), (*this)(-overhang,0), cv::Size2i(filterSize, filterSize), sigmaBegin, sigmaBegin, BORDER_REPLICATE);
		}

		// the first octave
		double sigma;
		for (int j=-overhang+1; j<scales+overhang; ++j) {
			sigma = sigmaDiff(j-1, 1.0);
			filterSize = 2*(int)(filterPrecision*sigma)+1;
			// iteratively blur the image
			cv::GaussianBlur((*this)(j-1,0), (*this)(j,0), cv::Size2i(filterSize, filterSize), sigma, sigma, BORDER_REPLICATE);
		}

		// the other octaves
		for (int i=1; i<octaves; ++i) {
			// in image convolved with sigma = 2*sB, sampling frequency can be halved
			if (overhang==0) {
				// there is not jet an image that was smoothed by one octave, so compute it
				cv::Mat_<_Tp> tmp;
				sigma = sigmaDiff(scales-1, 1.0);
				filterSize = 2*(int)(filterPrecision*sigma)+1;
				// iteratively blur the image
				cv::GaussianBlur((*this)(scales-1,i-1), tmp, cv::Size2i(filterSize, filterSize), sigma, sigma, BORDER_REPLICATE);
				cv::resize(tmp, (*this)(0,i), cv::Size2i(tmp.cols/2, tmp.rows/2), 0.0, 0.0, INTER_LINEAR);
			}
			else {
				cv::resize((*this)(scales-overhang,i-1), (*this)(-overhang,i), cv::Size2i((*this)(-overhang+scales,i-1).cols/2, (*this)(-overhang+scales,i-1).rows/2), 0.0, 0.0, INTER_LINEAR);
			}

			for (int j=-overhang+1; j<scales+overhang; ++j) {
				sigma = sigmaDiff(j-1, 1.0);
				filterSize = 2*(int)(filterPrecision*sigma)+1;
				cv::GaussianBlur((*this)(j-1,i), (*this)(j,i), cv::Size2i(filterSize, filterSize), sigma, sigma, BORDER_REPLICATE);
			}
		}
	}

	// number of scales in each octave
	const int scales;
	// number of octaves
	// sigma is doubled from one octave to the next, while the resolution is halfed
	const int octaves;
	// how many additional scales should be computed at the octave resolution?
	const int overhang;
	// pre-smoothing
	const double sigmaBegin;
	// sigma-factor for kernel size
	const double filterPrecision;

	// the gaussian-pyramid pixel data
	std::vector<std::vector<cv::Mat_<_Tp> > > pyramid;
};

} /* namespace cv */
#endif /* GPYR_H_ */
