//============================================================================
// Name        : ImageOperations.cpp
// Author      : D.A.Klein
// Version     : 1.1
// Copyright   : (C) University of Bonn
// Description : CoDi: Continuous-Distributions Saliency
// License     : GNU GPLv3 (see http://www.gnu.org/copyleft/gpl.html)
//============================================================================

/*
If you use this software or parts of it within scientific work, please cite
the original paper it was introduced:

@InProceedings{klein_dagm12,
  author = {Dominik A. Klein and Simone Frintrop},
  title = {{Salient Pattern Detection using $W_2$ on Multivariate Normal Distributions}},
  booktitle = {Proc.~of DAGM and OAGM},
  year = {2012},
  month = {Aug.},
  address = {Graz, Austria}
}
*/

#include "ImageOperations.h"

void cv::gaussianBoxBlur(cv::InputArray _src, cv::OutputArray _dst, double sigmaX, double sigmaY, int n, int borderType, double filterPrecision) {
	CV_Assert(sigmaX > 0.0 && n>0);

	if (sigmaY <= 0.0) sigmaY = sigmaX;

	cv::Mat src = _src.getMat();
    _dst.create( src.size(), src.type() );
    cv::Mat dst = _dst.getMat();

    // box filter sizes to approximate a gaussian
    int wx=1, wy=1;
    // how often the smaller filter is used
    int mx=n, my=n;
    // sigmas left for true Gaussian filtering
    double sX=sigmaX, sY=sigmaY;
    cv::Size2i ksize(std::max(2*(int)(filterPrecision*sX)+1, 3), std::max(2*(int)(filterPrecision*sY)+1, 3));

    // x-direction
	// compute the box filter sizes needed
	// next odd integer smaller than ideal size
	wx = (((((int)sqrt( (12.0*sigmaX*sigmaX) / n + 1.0))+1)/2)*2)-1;
	// how often we filter with wx
	mx = ceil(((12*sigmaX*sigmaX) - (n*wx*wx + 4*n*wx + 3*n))/(-4*wx - 4));
	// update sX
	sX = sqrt(sigmaX*sigmaX - ((-mx*(4*wx+4) + n*wx*wx + 4*n*wx + 3*n)/12.0));
	// check if box filtering pays off
	if (n < (ksize.width - std::max(2*(int)(filterPrecision*sX)+1, 3))) {
		// yes
		ksize.width = std::max(2*(int)(filterPrecision*sX)+1, 3);
	}
	else {
		// no
		wx = 1;
		mx = n;
	}

	// y-direction
    wy = (((((int)sqrt( (12.0*sigmaY*sigmaY) / n + 1.0))+1)/2)*2)-1;
    my = ceil(((12*sigmaY*sigmaY) - (n*wy*wy + 4*n*wy + 3*n))/(-4*wy - 4));
	sY = sqrt(sigmaY*sigmaY - ((-my*(4*wy+4) + n*wy*wy + 4*n*wy + 3*n)/12.0));
	if (n < (ksize.height - std::max(2*(int)(filterPrecision*sY)+1, 3))) {
		ksize.height = std::max(2*(int)(filterPrecision*sY)+1, 3);
	}
	else {
		wy = 1;
		my = n;
	}

    // box filtering
    if (wx > 1 || wy > 1) {
    	const int sumType = CV_MAKETYPE(CV_64F, src.channels());

    	// first, do the filtering for the small filters
    	cv::Ptr<BaseRowFilter> rowFilter = getRowSumFilter(src.type(), sumType, wx);
    	cv::Ptr<BaseColumnFilter> columnFilter = getColumnSumFilter(sumType, dst.type(), wy, -1, 1./(wx*wy));
    	cv::Ptr<FilterEngine> f = new FilterEngine(Ptr<BaseFilter>(0), rowFilter, columnFilter, src.type(), dst.type(), sumType, borderType);
    	f->apply( src, dst );
	    for (int i=1; i<std::min(mx,my); ++i)
	    	f->apply( dst, dst );

	    // continue with a bigger filter in one dimension
	    if (mx != my) {
	    	if (my > mx) {
	    		rowFilter = getRowSumFilter(src.type(), sumType, wx+2);
	    		columnFilter = getColumnSumFilter(sumType, dst.type(), wy, -1, 1./((wx+2)*wy));
	    		f = new FilterEngine(Ptr<BaseFilter>(0), rowFilter, columnFilter, src.type(), dst.type(), sumType, borderType);
	    	    for (int i=mx; i<my; ++i)
	    	    	f->apply( dst, dst );
	    	}
	    	else {
	    		columnFilter = getColumnSumFilter(sumType, dst.type(), wy+2, -1, 1./(wx*(wy+2)));
	    		f = new FilterEngine(Ptr<BaseFilter>(0), rowFilter, columnFilter, src.type(), dst.type(), sumType, borderType);
	    	    for (int i=my; i<mx; ++i)
	    	    	f->apply( dst, dst );
	    	}
	    }

	    // convolve with the bigger filter in both dimensions
	    if (std::max(mx,my) < n) {
			rowFilter = getRowSumFilter(src.type(), sumType, wx+2);
			columnFilter = getColumnSumFilter(sumType, dst.type(), wy+2, -1, 1./((wx+2)*(wy+2)));
			f = new FilterEngine(Ptr<BaseFilter>(0), rowFilter, columnFilter, src.type(), dst.type(), sumType, borderType);
			for (int i=std::max(mx,my); i<n; ++i)
				f->apply( dst, dst );
	    }

	    // finally, apply (smaller) Gaussian filter
	    cv::GaussianBlur(dst, dst, ksize, sX, sY, borderType);
    }
    else {
    	// box filtering is not used
    	cv::GaussianBlur(src, dst, ksize, sX, sY, borderType);
    }
}
