//============================================================================
// Name        : main.cpp
// Author      : D.A.Klein
// Version     : 1.1
// Copyright   : (C) University of Bonn
// Description : CoDi: Continuous-Distributions Saliency
// License     : GNU GPLv3 (see http://www.gnu.org/copyleft/gpl.html)
//============================================================================

/*
If you use this software or parts of it within scientific work, please cite
the original paper it was introduced:

@InProceedings{klein_dagm12,
  author = {Dominik A. Klein and Simone Frintrop},
  title = {{Salient Pattern Detection using $W_2$ on Multivariate Normal Distributions}},
  booktitle = {Proc.~of DAGM and OAGM},
  year = {2012},
  month = {Aug.},
  address = {Graz, Austria}
}
*/

#define TP float

#include <opencv2/opencv.hpp>
#include <omp.h>
#include <sys/time.h>

#include "../codi/CoDi.h"

bool save_intensity = false;
bool save_color = false;
bool save_blended = false;

char* filename_out_prefix = NULL;
char* filename_in_config = NULL;
std::vector<char*> filename_in_images;

// prints usage commands on the command line
void printHelp() {
	std::cout << "Usage: BITS-C [options] [image_filename_1] [image_filename_2] ... \n";
	std::cout << "Valid options are:\n";
	std::cout << "-I : save (I)ntensity map.\n";
	std::cout << "-C : save (C)olor map.\n";
	std::cout << "-B : save saliency (B)lended image.\n";
	std::cout << "-o [prefix]: output file(s) prefix. The output filename is [prefix][image_filename_$n]_[maptype].png.\n";
	std::cout << "-c [config]: configuration file. If omitted, standard config is used.\n";
	std::cout << "-h : print this information.\n";
}

// parses the given command line parameters and options, changes global variables
void parseCommandline(int argc, char **argv) {
    if (argc<2) {
    	std::cerr << "Wrong number of arguments. Call program with option -h for help.\n";
    	exit(1);
    }

    int o;
    while ((o = getopt (argc, argv, "IOCBo:c:h")) != -1) {
      switch (o)
        {
        case 'I': save_intensity = true; break;
        case 'C': save_color = true; break;
        case 'B': save_blended = true; break;
        case 'o': filename_out_prefix = optarg; break;
        case 'c': filename_in_config = optarg; break;
        case 'h': printHelp(); break;
        case '?':
        	if (optopt == 'c' || optopt == 'o')
        		std::cerr << "Option -" << optopt << " requires an argument.\n";
        	else if (isprint (optopt))
        		std::cerr << "Unknown option '-" << optopt << "'.\n Use option -h for a list of valid options.";
        	else
        		std::cerr << "Unknown option character '" << optopt << "'.\n Use option -h for a list of valid options.";
        	exit(1); break;
        default: exit(1); break;
        }
    }

    for (int i = optind; i < argc; i++)
    	filename_in_images.push_back(argv[i]);

    if (filename_in_images.size() == 0) {
    	std::cerr << "At least one input image must be given. Call program with option -h for help.\n";
    	exit(1);
    }
}

// returns the difference x-y in seconds
double operator- (const timeval& x, const timeval& y) {
	return (x.tv_sec - y.tv_sec) + 0.000001*(x.tv_usec - y.tv_usec);
}


int main(int argc, char **argv) {

#ifdef DEBUG
	omp_set_num_threads(1);
#else
	omp_set_num_threads(omp_get_num_procs());
#endif

	// get chosen options
	parseCommandline(argc, argv);

	// create cfg object with standard parameters
	cv::CoDicfg cfg;
	// if config file is given, read and use it
	if (filename_in_config) {
		std::ifstream conf_file(filename_in_config);
		if (conf_file.good()) {
			boost::archive::xml_iarchive ia(conf_file);
			ia >> BOOST_SERIALIZATION_NVP(cfg);
			conf_file.close();
		}
		else {
			std::cout << "Unable to open config file: " << filename_in_config << std::endl;
			return 1;
		}
	}

	// create CoDi with cfg
	cv::CoDi<TP> CoDi(cfg);
	// the image object for reading files, usually CV_8UC3 format
	cv::Mat input_img;
	// the image for converting to what is used in CoDi
	cv::Mat_<cv::Vec<TP, 3> > color_img;
	// storage for the results
	cv::Mat_<TP> result;
	// min and max to be used for range normalization to [0;1]
    double min,max;
    char filename_output[2048];
    // for timing statistics
    double comp_time = 0.0;
    int comp_images = 0;
    timeval tic, toc, start, stop;
    // start time measurement for the remaining time estimate
    gettimeofday(&tic, 0);
    int n=1, total=filename_in_images.size();
	std::vector<char*>::const_iterator citer = filename_in_images.begin();
	while (citer != filename_in_images.end()) {
		// load file
		input_img = cv::imread(*citer, CV_LOAD_IMAGE_COLOR);
		// convert to TP
		input_img.convertTo(color_img, cv::DataType<cv::Vec<TP, 3> >::type);
		// norm range to [0;1]
		color_img *= (1./255.);
		// start time measurement for saliency computation
		gettimeofday(&start, 0);
		// compute saliency
		CoDi.computeFrom(color_img);
		// stop time measurement
		gettimeofday(&stop, 0);
		// add up for average computation time
		comp_time += (stop - start);
		comp_images++;
		// save output in range [0,255]
		// saliency map:
		result = CoDi.getSalMap();
		// norm range to [0;1]
		cv::minMaxLoc(result, &min, &max);
		result -= min;
		result *= (255.0/(max-min));
		// store as file
		if (filename_out_prefix)
			strcpy(filename_output, filename_out_prefix);
		else
			strcpy(filename_output, "");
		strcat(filename_output, *citer);
		strcat(filename_output, "_saliency.png");
		cv::imwrite(filename_output, result);

		// save image blended by saliency value
		if (save_blended) {
			cv::Mat_<cv::Vec<TP, 4> > blend_img;
			cv::Mat mv[2] = { color_img, result};
			cv::merge(mv, 2, blend_img);

			if (filename_out_prefix)
				strcpy(filename_output, filename_out_prefix);
			else
				strcpy(filename_output, "");
			strcat(filename_output, *citer);
			strcat(filename_output, "_blended.png");
			cv::imwrite(filename_output, blend_img);
		}

		// save intensity conspicuity
		if (save_intensity) {
			result = CoDi.getIntensityConsp();
			cv::minMaxLoc(result, &min, &max);
			result -= min;
			result *= (255.0/(max-min));
			if (filename_out_prefix)
				strcpy(filename_output, filename_out_prefix);
			else
				strcpy(filename_output, "");
			strcat(filename_output, *citer);
			strcat(filename_output, "_intensity.png");
			cv::imwrite(filename_output, result);
		}

		// save color conspicuity
		if (save_color) {
			result = CoDi.getColorConsp();
			cv::minMaxLoc(result, &min, &max);
			result -= min;
			result *= (255.0/(max-min));
			if (filename_out_prefix)
				strcpy(filename_output, filename_out_prefix);
			else
				strcpy(filename_output, "");
			strcat(filename_output, *citer);
			strcat(filename_output, "_color.png");
			cv::imwrite(filename_output, result);
		}

		// elapsed time is toc - tic
		gettimeofday(&toc, 0);
		// approx. remaining time
		double est_time = ((toc - tic)/(double)n) * (total-n);
		std::cout << "Stored results for image " << n << " of " << total << " (" << *citer << "), ~" << est_time << " seconds left.\n";

		++citer;
		++n;
	}

	std::cout << comp_images << " images computed in " << comp_time << " seconds.\n" << comp_time/(double)comp_images << " seconds/image.\n";

	// done
	return 0;
}
